/*****************************************************************************
 * FIPS140-2_HMAC_BF70x.c
 *****************************************************************************/

#include "adi_initialize.h"
#include "FIPS140-2_HMAC_BF70x.h"
#include <blackfin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


cd_t cmdDes[4];
rd_t resDes[4];
sa_t saRecord[4];
state_t saState[4];


bool run_HMAC_validation(char *);


int main(void)
{
	/**
	 * Initialize managed drivers and/or services that have been added to 
	 * the project.
	 * @return zero on success 
	 */
	adi_initComponents();
	
	/* run HMAC */
	run_HMAC_validation("../src/HMAC.req");

	return 0;
}


uint8_t inMsg[128*2];    // 1024 bits (max)

bool run_HMAC_validation(char *filename)
{

	FILE *in_fp, *out_fp;
	uint32_t cnt;
	char out_filename[128];
	int i;
	shaVer hsh;
	uint32_t macOutput[32*2];// 512 bits (max) - add padding for stack issues
	uint32_t key[36];

	uint32_t klen;
	uint32_t tlen;
	uint32_t count;
	uint32_t msglen=128;

	/* reset engine and release */
	*pREG_PKTE0_CFG = 3;
	*pREG_PKTE0_CFG = 0;
	*pREG_PKTE0_CFG = 0x300;           // configure for Autonomous Ring Mode


	*pREG_PKTE0_RING_CFG = 2;          // three descriptors ring
	*pREG_PKTE0_CDRBASE_ADDR = (uint32_t)(&cmdDes);
	*pREG_PKTE0_RDRBASE_ADDR = (uint32_t)(&resDes);

	in_fp = fopen(filename, "r");
	if(in_fp == NULL){
		printf("can't open input file\n");
		exit(0);
	}


	out_fp = fopen("HMAC.rsp", "w");

	if(out_fp == NULL) {
		printf("can't open one or all of the output files\n");
		exit(0);
	}



	/* run through test file */
	while(!feof(in_fp)){
		/* get next test vector data from file */
		read_test_vector_data(&count, &klen, &tlen, key, inMsg, in_fp, out_fp, &hsh);

		/* run test vector through engine */
		run_HMAC(count, klen, tlen, key, inMsg, macOutput, hsh, msglen);

		/* output result to result file */
		output_test_result(macOutput, tlen, out_fp);

		free(inMsg);

	}

	fclose(in_fp);
	fclose(out_fp);

	return true;
}

char line[512] = "";
uint32_t read_test_vector_data(uint32_t *count, uint32_t *kl, uint32_t *tl, uint32_t *keybuf, uint32_t *msg, FILE *fp, FILE *ofp, shaVer *hash)
{

	char *inMsgLine;
	uint32_t len;
	int i, scanfRes, digestSz;
	char *resFgets;
	long whereAmI;


	if(fp == NULL){
		printf("no file to read\n");
		exit(0);
	}

	while(strncmp(line, "Count", 5)!=0){
		resFgets = fgets(line, sizeof(line), fp);
		fputs(line, ofp);

		if(strncmp(line, "[L=", 3)==0){
			digestSz = strtoul(&line[3], 0, 10);
			printf("L=%d\n", digestSz);

			if(digestSz == 20)
				*hash = SHA1;
			else if(digestSz == 28)
				*hash = SHA224;
			else if(digestSz == 32)
				*hash = SHA256;

		}


		if(resFgets == NULL || feof(fp))  // return if we reach end of file
			return 0;
	}

	if(strncmp(line, "Count", 5) == 0){
		*count = strtoul(&line[8], 0, 10);
		printf("count=%d\n", *count);
	}

	resFgets = fgets(line, sizeof(line), fp);
	while(line[0] != '\n' && resFgets != NULL  &&  !feof(fp) ){
		if(strncmp(line, "Klen", 4) == 0){
			*kl = strtoul(&line[7], 0, 10);
			fputs(line, ofp);
		}

		if(strncmp(line, "Tlen", 4) == 0){
			*tl = strtoul(&line[7], 0, 10);
			fputs(line, ofp);
		}

		if(strncmp(line, "Key", 3) == 0){
			fputs(line, ofp);

			for(i=0; i<*kl; i++)
				sscanf(&line[6 + i *2], "%02x", &((uint8_t *)keybuf)[i]);

		}

		if(strncmp(line, "Msg", 3) == 0){
			fputs(line, ofp);
			for(i=0;i<128; i++){
				sscanf(&line[6 + i *2], "%02x", &((uint8_t *)msg)[i]);
			}
			printf("%d msg bytes scanned\n", i);

		}

		resFgets = fgets(line, sizeof(line), fp);

	}

	return 0;

}

void run_HMAC(uint32_t count, uint32_t kl, uint32_t tl, uint32_t *keybuf, uint32_t *msg, uint32_t *macOut, shaVer hash, uint32_t msglen)
{

	int i, ringPtr, digestLen;
	uint8_t *kbuf = (uint8_t *)keybuf, t;
	cd_t *cmdDescriptor;
	uint8_t *k0ipad, *k0opad;



	if(hash == SHA1)
		digestLen =  5;  // # of words
	else if(hash == SHA224)
		digestLen = 7;
	else if(hash == SHA256)
		digestLen = 8;


	/* 64 bytes = 512 bits, block size of SHA-1 */

	/* allocate memory for k0ipad and k0opad */
	k0ipad = malloc(64);
	k0opad = malloc(64);

	if(k0ipad == NULL || k0opad == NULL){
		printf("could not allocate memory for k0ipad or k0opad\n");
		exit(0);
	}

	/* if key is greater than block size, hash it */
	if(kl > 64){
		/* run hash on key to produce L byte string and then pad with B-L zeros */

		cmdDescriptor = &cmdDes[0];

		/* set up command descriptor */
		cmdDescriptor->CTRL_STAT = (0x1 | (CTRL_STAT_HASH_FINAL<<4));  // transfer ownership from host to PE of the desc

		cmdDescriptor->SRC_ADDR  = (unsigned long)keybuf;
		cmdDescriptor->DEST_ADDR = (unsigned long)keybuf;  // result will be in STATE

		cmdDescriptor->SA_ADDR = (unsigned long)(&saRecord[0]);
		cmdDescriptor->SA_STATE_ADDR = (unsigned long)(&saState[0]);
		cmdDescriptor->ARC4_STATE_ADDR = 0;
		cmdDescriptor->USERID = 0;
		cmdDescriptor->LENGTH = (kl) | 0x00400000;


		/* configure SA CMD regs */
		saRecord[0].CMD_1 = (SA_CMD1_CPYDIGEST << 0) | \
					(SA_CMD1_CPYPAD  << 3);
		saRecord[0].CMD_0 = (SA_CMD0_OPCODE_HASH<<0) | \
				(SA_CMD0_DIR_IN<<3) |               \
				(SA_CMD0_PAD_ZERO<<7) |             \
				(SA_CMD0_CIPHER_NULL<<8) |          \
				(hash<<12)  |                       \
				(digestLen<<20) |                   \
				(SA_CMD0_HASHSRC_NOLOAD<<26);


		/* start processing */
		*pREG_PKTE0_CDSC_INCR = 1;

		//while(*pREG_PKTE0_RDSC_CNT == 0)
		while(*pREG_PKTE0_RDSC_CNT != 1)
			for(i=0; i<100; i++);

		/* update result desc ring */
		*pREG_PKTE0_RDSC_DECR = 1;

		/* since this is not done all the time, use 1st desc in ring and then reset */
		*pREG_PKTE0_CFG = 0x302;           // configure for Autonomous Ring Mode and reset ring
		*pREG_PKTE0_CFG = 0x300;           // release from reset


		/* reconfigure the rings */
		*pREG_PKTE0_RING_CFG = 2;          // three descriptors ring
		*pREG_PKTE0_CDRBASE_ADDR = (uint32_t)(&cmdDes);
		*pREG_PKTE0_RDRBASE_ADDR = (uint32_t)(&resDes);


		/* byte swap output.  The new K0 is byte-swapped of hash(key) */
		for(i=0; i<(digestLen * 4); i+=4){
			t = kbuf[i];
			kbuf[i] = kbuf[i+3];
			kbuf[i+3] = t;
			t = kbuf[i+1];
			kbuf[i+1] = kbuf[i+2];
			kbuf[i+2] = t;
		}


		/* the hashed key is the hash digest length.  pad to block size */
		for(i=(digestLen * 4); i<64; i++)
			kbuf[i] = 0;

	}

	/* copy key into new buffers */
	if(kl <= 64){
		for(i=0; i<kl; i++){
			k0ipad[i] = ((uint8_t *)kbuf)[i];
			k0opad[i] = ((uint8_t *)kbuf)[i];
		}

		/* if key is less than block size, pad end with zeros */
		if(kl < 64)
			for(i=kl; i<64; i++){
				k0ipad[i] = 0;
				k0opad[i] = 0;
			}
	}

	/* if kl is greater than 512bits, the new key is already hashed and padded to 512 bits */
	if(kl > 64)
		for(i=0; i<64; i++){
			k0ipad[i] = ((uint8_t *)kbuf)[i];
			k0opad[i] = ((uint8_t *)kbuf)[i];
		}


	/* perform the XOR to obtain k0ipad and k0opad */
	for(i=0; i<64; i++){
		k0ipad[i] ^= 0x36;
		k0opad[i] ^= 0x5c;
	}


	/*****************/
	/* H(K0 ^ ipad)  */
	/*****************/

	//ringPtr = *pREG_PKTE0_RING_PTR;
	cmdDescriptor = &cmdDes[0];

	/* set up command descriptor */
	cmdDescriptor->CTRL_STAT = 0x1;                    // transfer ownership from host to PE of the desc

	cmdDescriptor->SRC_ADDR  = (unsigned long)k0ipad;
	cmdDescriptor->DEST_ADDR = (unsigned long)macOut;  // result will be in STATE

	cmdDescriptor->SA_ADDR = (unsigned long)(&saRecord[0]);
	cmdDescriptor->SA_STATE_ADDR = (unsigned long)(&saState[0]);
	cmdDescriptor->ARC4_STATE_ADDR = 0;
	cmdDescriptor->USERID = 0;
	cmdDescriptor->LENGTH = (64) | 0x00400000;


	/* configure SA CMD regs */
	saRecord[0].CMD_1 = 0;
	saRecord[0].CMD_0 = (SA_CMD0_OPCODE_HASH<<0) | \
			(SA_CMD0_DIR_IN<<3) |               \
			(SA_CMD0_PAD_ZERO<<7) |             \
			(SA_CMD0_CIPHER_NULL<<8) |          \
			(hash<<12)  |                       \
			(digestLen<<20) |         \
			(SA_CMD0_HASHSRC_NOLOAD<<26) |      \
			(SA_CMD0_SVHASH_YES<<29) ;

	/****************************************************************************/
	/* H(K0 ^ opad)                                                             */
	/* The output will be the new "starting" hash to the final hash calculation */
	/****************************************************************************/

	cmdDescriptor = &cmdDes[1];

	/* set up command descriptor */
	cmdDescriptor->CTRL_STAT = 0x1;                    // transfer ownership from host to PE of the desc

	cmdDescriptor->SRC_ADDR  = (unsigned long)k0opad;
	cmdDescriptor->DEST_ADDR = (unsigned long)macOut;  // result will be in STATE

	cmdDescriptor->SA_ADDR = (unsigned long)(&saRecord[1]);
	cmdDescriptor->SA_STATE_ADDR = (unsigned long)(&saState[1]);
	cmdDescriptor->ARC4_STATE_ADDR = 0;
	cmdDescriptor->USERID = 0;
	cmdDescriptor->LENGTH = (64) | 0x00400000;


	/* configure SA CMD regs */
	saRecord[1].CMD_1 = 0;
	saRecord[1].CMD_0 = (SA_CMD0_OPCODE_HASH<<0) | \
			(SA_CMD0_DIR_IN<<3) |               \
			(SA_CMD0_PAD_ZERO<<7) |             \
			(SA_CMD0_CIPHER_NULL<<8) |          \
			(hash<<12)  |                       \
			(digestLen<<20) |         \
			(SA_CMD0_HASHSRC_NOLOAD<<26) |      \
			(SA_CMD0_SVHASH_YES<<29) ;

	/* Must calculate the inner and outer hash first */

	/* start processing */
	*pREG_PKTE0_CDSC_INCR = 2;

	//while(*pREG_PKTE0_RDSC_CNT == 0)
	while(*pREG_PKTE0_RDSC_CNT != 2)
		for(i=0; i<100; i++);

	/* update result desc ring */
	*pREG_PKTE0_RDSC_DECR = 2;




	/**********************************************/
	/* Final MAC calculation                      */
	/* H( (K0 ^ opad) || H( (K0 ^ ipad) || msg) ) */
	/**********************************************/

	cmdDescriptor = &cmdDes[2];

	/* set up command descriptor */
	cmdDescriptor->CTRL_STAT = (0x1 | (CTRL_STAT_HASH_FINAL<<4));                    // transfer ownership from host to PE of the desc

	cmdDescriptor->SRC_ADDR  = (unsigned long)msg;
	cmdDescriptor->DEST_ADDR = (unsigned long)macOut;  // result will be in STATE

	cmdDescriptor->SA_ADDR = (unsigned long)(&saRecord[2]);
	cmdDescriptor->SA_STATE_ADDR = (unsigned long)(&saState[2]);
	cmdDescriptor->ARC4_STATE_ADDR = 0;
	cmdDescriptor->USERID = 0;
	cmdDescriptor->LENGTH = (msglen) | 0x00400000;  // assuming all tests are 128byte messages

	/* load pre-calculated inner and outer digest */
	saRecord[2].IDIGEST_0 = saState[0].STATE_IDIGEST_0;
	saRecord[2].IDIGEST_1 = saState[0].STATE_IDIGEST_1;
	saRecord[2].IDIGEST_2 = saState[0].STATE_IDIGEST_2;
	saRecord[2].IDIGEST_3 = saState[0].STATE_IDIGEST_3;
	saRecord[2].IDIGEST_4 = saState[0].STATE_IDIGEST_4;
	saRecord[2].IDIGEST_5 = saState[0].STATE_IDIGEST_5;
	saRecord[2].IDIGEST_6 = saState[0].STATE_IDIGEST_6;
	saRecord[2].IDIGEST_7 = saState[0].STATE_IDIGEST_7;

	saRecord[2].ODIGEST_0 = saState[1].STATE_IDIGEST_0;
	saRecord[2].ODIGEST_1 = saState[1].STATE_IDIGEST_1;
	saRecord[2].ODIGEST_2 = saState[1].STATE_IDIGEST_2;
	saRecord[2].ODIGEST_3 = saState[1].STATE_IDIGEST_3;
	saRecord[2].ODIGEST_4 = saState[1].STATE_IDIGEST_4;
	saRecord[2].ODIGEST_5 = saState[1].STATE_IDIGEST_5;
	saRecord[2].ODIGEST_6 = saState[1].STATE_IDIGEST_6;
	saRecord[2].ODIGEST_7 = saState[1].STATE_IDIGEST_7;


	/* configure SA CMD regs */
	saRecord[2].CMD_1 = (SA_CMD1_CPYDIGEST << 0) | \
			(SA_CMD1_CPYPAD  << 3) |               \
			(SA_CMD1_HMAC << 12);
	saRecord[2].CMD_0 = (SA_CMD0_OPCODE_HASH<<0) | \
			(SA_CMD0_DIR_IN<<3) |               \
			(SA_CMD0_PAD_ZERO<<7) |             \
			(SA_CMD0_CIPHER_NULL<<8) |          \
			(hash<<12)  |                       \
			(digestLen<<20) |         \
			(SA_CMD0_HASHSRC_FRMSA<<26) |       \
	        (SA_CMD0_SVHASH_YES<<29) ;



	/* start processing */
	*pREG_PKTE0_CDSC_INCR = 1;

	while(*pREG_PKTE0_RDSC_CNT != 1)
		for(i=0; i<100; i++);

	/* update result desc ring */
	*pREG_PKTE0_RDSC_DECR = 1;


	free(k0ipad);
	free(k0opad);
}



void output_test_result(uint32_t *output, uint32_t tl, FILE *fp)
{
	int i, j;
	uint8_t *out = (uint8_t *)output;


	/* output is byteswapped */
	fprintf(fp, "Mac = ");
	for(i=0; i<(tl&0xfffffffc); i+=4){
		fprintf(fp, "%02x", out[i+3]);
		fprintf(fp, "%02x", out[i+2]);
		fprintf(fp, "%02x", out[i+1]);
		fprintf(fp, "%02x", out[i+0]);
	}


	for(j=0; j<(tl&3); j++)
		fprintf(fp, "%02x", out[i+3-j]);

	fprintf(fp, "\n\n");



}



